import java.io.PrintWriter
import scala.virtualization.lms.common._
import scala.virtualization.lms.internal.Scheduling
import scala.virtualization.lms.util.OverloadHack

trait Example4Ops extends Base with OverloadHack {
  object Matrix {
    def apply[T:Manifest](x: Rep[Int], y: Rep[Int]) = matrix_new[T](x,y)
  }

  def infix_apply[T:Manifest](x: Rep[Matrix[T]], m: Rep[Int], n: Rep[Int]) = matrix_apply(x,m,n)
  def infix_update[T:Manifest](x: Rep[Matrix[T]], m: Rep[Int], n: Rep[Int], y: Rep[T]) = matrix_update(x,m,n,y)
  def infix_*[T](x: Rep[Matrix[T]], y: Rep[Matrix[T]])(implicit mA: Manifest[T], o: Overloaded1) = matrix_times(x,y)
  def infix_*[T](x: Rep[Matrix[T]], y: Rep[Int])(implicit mT: Manifest[T], o: Overloaded2) = matrix_times_scalar(x,y)
  def println[T:Manifest](x: Rep[Matrix[T]]) = matrix_print(x)

  def matrix_new[T:Manifest](x: Rep[Int], y: Rep[Int]): Rep[Matrix[T]]
  def matrix_apply[T:Manifest](x: Rep[Matrix[T]], m: Rep[Int], n: Rep[Int]): Rep[T]
  def matrix_update[T:Manifest](x: Rep[Matrix[T]], m: Rep[Int], n: Rep[Int], y: Rep[T]): Rep[Unit]
  def matrix_times[T:Manifest](x: Rep[Matrix[T]], y: Rep[Matrix[T]]): Rep[Matrix[T]]
  def matrix_times_scalar[T:Manifest](x: Rep[Matrix[T]], y: Rep[Int]): Rep[Matrix[T]]
  def matrix_print[T:Manifest](x: Rep[Matrix[T]]): Rep[Unit]
}

trait Example4OpsExp extends Example4Ops with EffectExp {
  case class MatrixNew[T:Manifest](x: Exp[Int], y: Exp[Int]) extends Def[Matrix[T]]
  case class MatrixApply[T:Manifest](x: Exp[Matrix[T]], m: Exp[Int], n: Exp[Int]) extends Def[T]
  case class MatrixUpdate[T:Manifest](x: Exp[Matrix[T]], m: Exp[Int], n: Exp[Int], y: Exp[T]) extends Def[Unit]
  case class MatrixTimes[T:Manifest](x: Exp[Matrix[T]], y: Exp[Matrix[T]]) extends Def[Matrix[T]]
  case class MatrixTimesScalar[T:Manifest](x: Exp[Matrix[T]], y: Exp[Int]) extends Def[Matrix[T]]
  case class MatrixPrint[T:Manifest](x: Exp[Matrix[T]]) extends Def[Unit]

  def matrix_new[T:Manifest](x: Exp[Int], y: Exp[Int]) = reflectMutable(MatrixNew[T](x,y))
  // unsafe! use this version and observe what happens in the generated code
  //def matrix_new[T:Manifest](x: Exp[Int], y: Exp[Int]) = MatrixNew[T](x,y)
  def matrix_apply[T:Manifest](x: Exp[Matrix[T]], m: Exp[Int], n: Exp[Int]) = MatrixApply(x,m,n)
  def matrix_update[T:Manifest](x: Exp[Matrix[T]], m: Exp[Int], n: Exp[Int], y: Exp[T]) = reflectWrite(x)(MatrixUpdate(x,m,n,y))
  // unsafe! use this version and observe what happens in the generated code
  //def matrix_update[T:Manifest](x: Exp[Matrix[T]], m: Exp[Int], n: Exp[Int], y: Exp[T]) = MatrixUpdate(x,m,n,y)
  def matrix_times[T:Manifest](x: Exp[Matrix[T]], y: Exp[Matrix[T]]) = MatrixTimes(x,y)
  def matrix_times_scalar[T:Manifest](x: Exp[Matrix[T]], y: Exp[Int]) = MatrixTimesScalar(x,y)
  def matrix_print[T:Manifest](x: Exp[Matrix[T]]) = reflectEffect(MatrixPrint(x))
  // unsafe! use this version and observe what happens in the generated code
  //def matrix_print[T:Manifest](x: Exp[Matrix[T]]) = MatrixPrint(x)
}

trait Example4ScalaGen extends ScalaGenEffect {
  val IR: Example4OpsExp
  import IR._

  override def emitNode(sym: Sym[Any], rhs: Def[Any])(implicit stream: PrintWriter): Unit = rhs match {
    case MatrixNew(m,n) => emitValDef(sym, "new MatrixImpl(" + quote(m) + "," + quote(n) + ")")
    case MatrixApply(x,m,n) => emitValDef(sym, quote(x) + ".apply(" + quote(m) + "," + quote(n) + ")")
    case MatrixUpdate(x,m,n,y) => emitValDef(sym, quote(x) + ".update(" + quote(m) + "," + quote(n) + "," + quote(y) + ")")
    case MatrixTimes(x,y) => emitValDef(sym, quote(x) + " * " + quote(y))
    case MatrixTimesScalar(x,y) => emitValDef(sym, quote(x) + " * " + quote(y))
    case MatrixPrint(x) => emitValDef(sym, "println(" + quote(x) + ")")
    case _ => super.emitNode(sym, rhs)
  }
}